/*
* Teekran.c
*
* Created: 10.08.2013 16:31:28
*  Author: Nicolas (www.DL1DOW.de)
*/

//LFUSE: 0x62
//HFUSE: 0xDD
//EFUSE: 0xFF

//#define DEBUG



#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdbool.h>

#include "led_multiplex.h"
#include "encoder_panasonic.h"
#include "ioMapping.h"
#include "keys.h"
#include "clock.h"
#include "servo.h"

#define SHORT_LONG 500
#define SLEEP_DELAY 1000
#define SERVO_DELAY_LIFT 6
#define SERVO_DELAY_DOWN 2

// Default-Werte: Richtung erkennbar, im erlaubten Bereich aber sehr wenig Hub.
uint16_t eeServoHigh EEMEM = 250; //200; 
uint16_t eeServoLow EEMEM = 125; // 150;

#define SERVO_MAX (F_CPU/8/454)  // ca. 3,6ms, empirisch mit Servo getestet
#define SERVO_MIN (F_CPU/8/1250) // 0,8ms, empirisch mit Servo getestet

typedef enum {tick,read,reset} sleeptimer_t;
uint8_t sleeptimer(sleeptimer_t command);


// Timer-Interruptroutine (Periode ca. 4ms)
//SIGNAL (SIG_OUTPUT_COMPARE2A)
ISR (TIMER2_COMPA_vect)
{
    FREI0_PORT |= (1<<FREI0);
    
    // LEDs multiplexen
    led_mpx_multiplex();

    // Drehgeber pollen
    encoder_get();
    
    // Taster pollen
    key_poll();
    
    // Uhr ticken lassen
    clock_tick();
    sleeptimer(tick);
    
    FREI0_PORT &= ~(1<<FREI0);
}

// Wird benoetigt fuer Wakeup ueber pin change interrupt
ISR (PCINT1_vect) {}


void sleep(void) 
{
    led_mpx_deinit();
    servo_off();
    
    // Pin change interrupt vorbereiten
    PCICR  |= (1 << PCIE1);
    PCMSK1 |= (1<<PCINT11);
    ACSR = 0x80;
    
    sei(); // Nur fuer alle Faelle
    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_enable();
    sleep_cpu();
    
    _delay_ms(10);
    
    // Pin change interrupts wieder sperren
    PCMSK1 &= ~(1<<PCINT11);
    
    led_mpx_init();
    servo_init(0);
}


// Mini-Clock-Funktion als Schlaftimer
uint8_t sleeptimer(sleeptimer_t command) 
{
    static uint16_t timer = SLEEP_DELAY;
    switch (command) {
        case tick:
            if (timer>0) timer--;
            break;
        case read:
            break;
        case reset:
            timer = SLEEP_DELAY;
    }
    return (timer==0);
}


// Einstellungen: Obere und untere Grenze des Servos
void setup(void) 
{
    enum {tohigh,high,tolow,low,last} setup_mode = tohigh;

    uint16_t shi = eeprom_read_word(&eeServoHigh);
    uint16_t slo = eeprom_read_word(&eeServoLow);
    uint8_t enc_this = encoder_get();
    uint8_t enc_last = enc_this;
    uint16_t *enc_var =  &shi;
    
    uint16_t kt = 0;
    
    servo_on();
    
    while(1) {
        // Tastendruck interpretieren: Ausgang oder weiterschalten
        kt = key_time();
        if(kt) {
            if (kt > SHORT_LONG) {
                break;
            } 
            else {
                setup_mode++;
                if (setup_mode==last) setup_mode = 0;
            }
        } // if(kt)
        
        // Drehgeber auswerten (nur Aenderungen)
        enc_this = encoder_get();
        *enc_var += (int8_t) (enc_this-enc_last);
        enc_last = enc_this;
        
        switch (setup_mode) {
            case tohigh: // High-Stellung einstellen (Start)
                led_mpx_putany((char[]) {0x00,LED_CHAR_I,LED_CHAR_H});
                enc_this = encoder_get();
                enc_last = enc_this;
                enc_var = &shi;
                setup_mode = high;
                /* FALLTHRU */
            case high: // High-Stellung einstellen
                #ifdef DEBUG
                    led_mpx_putnum(shi);
                #endif
            
                if(shi>SERVO_MAX)
                    shi=SERVO_MAX;
                if(shi<SERVO_MIN)
                    shi = SERVO_MIN;
            
                OCR1A = shi;
                break;
            
            case tolow: // Low-Stellung einstellen (Start)
                led_mpx_putany((char[]) {0x00,LED_CHAR_O,LED_CHAR_L});
                enc_this = encoder_get();
                enc_last = enc_this;
                enc_var = &slo;
                setup_mode = low;
                /* FALLTHRU */
            case low: // Low-Stellung einstellen
                #ifdef DEBUG
                    led_mpx_putnum(slo);
                #endif            
            
                if(slo>SERVO_MAX)
                    slo=SERVO_MAX;
                if(slo<SERVO_MIN)
                    slo = SERVO_MIN;  
                            
                OCR1A = slo;
        }
    } // while(1)
    
    eeprom_write_word(&eeServoHigh, shi);
    eeprom_write_word(&eeServoLow, slo);
}


int main(void) 
{
    led_mpx_init();
    encoder_init();
    key_init();
    servo_init(eeprom_read_word(&eeServoHigh));
    
    // Leere Pins
    FREI0_PORT |= (1<<FREI0);
    FREI1_PORT |= (1<<FREI1);
    FREI2_PORT |= (1<<FREI2);
    FREI3_PORT |= (1<<FREI3);
    FREI0_DDR  |= (1<<FREI0);
    FREI1_DDR  |= (1<<FREI1);
    
    // Initialisierung Uhr/Sekundenzhler: (250Hz) (S. 168)
    TCCR2A = (1<<WGM21) | (0<<WGM20);
    #if F_CPU==500000L
        // Prescaler/Vorteiler = 8 clear on compare
        TCCR2B = (0<<WGM22) | (0<<CS22) | (1<<CS21) | (0 << CS20); 
        OCR2A = 250;  // Output Compare Register (8 Bit)
    #elif F_CPU==1000000L
        // Prescaler/Vorteiler = 32 clear on compare
        TCCR2B = (0<<WGM22) | (0<<CS22) | (1<<CS21) | (1 << CS20); 
        OCR2A = 125; // Output Compare Register (8 Bit)
    #else
        #error "unsupported F_CPU"
    #endif
    
    
    
    TCNT2 = 0;            // Timer startet mit 0
    TIMSK2 = 1<<OCIE2A;   // beim Vergleichswert Compare Match
    
    sei();

    int16_t counter = 0;
    int16_t helpcounter = 0;
    
    // Hilfswerte Drehgeber
    uint8_t enc_this = encoder_get();
    uint8_t enc_last = enc_this;
    int8_t enc_diff;
    
    enum {mclockrun,mclockset,mlift,mdown,mlast} mode = mclockrun;
    
    while(1) {
        // Drehgeber auswerten (nur Aenderungen)
        enc_this = encoder_get();
        enc_diff = (enc_this-enc_last);
        enc_last = enc_this;
        if (enc_diff) sleeptimer(reset);
        
        counter += 10*enc_diff;
        if (counter<0) counter=0;
        
        // Modi und/oder Konfiguration umschalten
        // Tastendruck
        bool iskeyaction = false;
        uint16_t p = key_time();
        if (p) {
            if (p<SHORT_LONG) {
                iskeyaction = true;
            }            
            else {
                setup();
                sleeptimer(reset);
                mode = mclockset;
            }
        }
        
        switch (mode) {
            case mclockset:
                // Uhr einstellen / Ruhelage
                
                // hier darf der Timer endlich mal...
            
                if (iskeyaction && counter) mode = mdown;            
                servo_off();
                
                led_mpx_blink(1);
                clock_stop();
                clock_set(counter);
                
                // wenn nichts passiert schlafen legen
                if (sleeptimer(read)) {
                    sleep();
                    sleeptimer(reset);
                }
            
                break;
            case mclockrun:
                // Uhr laeuft, Kran abgesenkt
                sleeptimer(reset);   
                             
                if (iskeyaction) mode = mlift;                
                servo_off();
                
                led_mpx_blink(0);
                clock_start();
                counter = clock_get();

                // Uhr abgelaufen, Rueckkehr in Ruhelage
                if (counter == 0) mode = mlift;
                break;
                
            case mlift:
                // Kran anheben 
                // - Bei Tastendruck sofort wieder senken
                // - Bei normalem Ende in "Uhr-Stellen"-Bereitschaft
                sleeptimer(reset);                 
                if (iskeyaction) mode = mdown;
                servo_on();


                {
                    bool isdone;
                    uint16_t tmp = eeprom_read_word(&eeServoHigh);
                    isdone = servo_move(tmp);
                    if(isdone) mode = mclockset;
                }
                
                // Dem Servo noch etwas mehr Zeit geben
                _delay_ms(SERVO_DELAY_LIFT);
                
                break;
                
            case mdown:
                // Kran absenken 
                // - Bei Tastendruck sofort wieder heben    
                // - Bei normalem Ende in "Uhr-Laeuft"-Modus
                sleeptimer(reset);
                if (iskeyaction) mode = mlift;
                servo_on();
                               
                {
                    bool isdone;
                    uint16_t tmp = eeprom_read_word(&eeServoLow);
                    isdone = servo_move(tmp);
                    if(isdone) mode = mclockrun;
                }
                
                // Dem Servo noch etwas mehr Zeit geben
                _delay_ms(SERVO_DELAY_DOWN); 
                              
        }
        
        led_mpx_putnum(counter);
        
        _delay_ms(5); // Zykluszeit etwas strecken
    } // while(1)
} // main





